/*
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */

package com.facebook.react.views.scroll;

import android.animation.Animator;
import android.animation.ObjectAnimator;
import android.animation.PropertyValuesHolder;
import android.animation.ValueAnimator;
import android.content.Context;
import android.graphics.Point;
import android.view.View;
import android.view.ViewGroup;
import android.widget.OverScroller;
import androidx.annotation.Nullable;
import com.facebook.common.logging.FLog;
import com.facebook.react.bridge.JSApplicationIllegalArgumentException;
import com.facebook.react.bridge.ReactContext;
import com.facebook.react.bridge.WritableMap;
import com.facebook.react.bridge.WritableNativeMap;
import com.facebook.react.common.build.ReactBuildConfig;
import com.facebook.react.uimanager.FabricViewStateManager;
import com.facebook.react.uimanager.PixelUtil;
import com.facebook.react.uimanager.UIManagerHelper;
import com.facebook.react.uimanager.common.UIManagerType;
import com.facebook.react.uimanager.common.ViewUtil;
import java.util.Collections;
import java.util.Set;
import java.util.WeakHashMap;

/** Helper class that deals with emitting Scroll Events. */
public class ReactScrollViewHelper {
  private static String TAG = ReactHorizontalScrollView.class.getSimpleName();
  private static boolean DEBUG_MODE = false && ReactBuildConfig.DEBUG;
  private static final String CONTENT_OFFSET_LEFT = "contentOffsetLeft";
  private static final String CONTENT_OFFSET_TOP = "contentOffsetTop";
  private static final String SCROLL_AWAY_PADDING_TOP = "scrollAwayPaddingTop";

  public static final long MOMENTUM_DELAY = 20;
  public static final String OVER_SCROLL_ALWAYS = "always";
  public static final String AUTO = "auto";
  public static final String OVER_SCROLL_NEVER = "never";

  public static final int SNAP_ALIGNMENT_DISABLED = 0;
  public static final int SNAP_ALIGNMENT_START = 1;
  public static final int SNAP_ALIGNMENT_CENTER = 2;
  public static final int SNAP_ALIGNMENT_END = 3;

  private static @Nullable ValueAnimator sScrollAnimator;

  public interface ScrollListener {
    void onScroll(
        ViewGroup scrollView, ScrollEventType scrollEventType, float xVelocity, float yVelocity);

    void onLayout(ViewGroup scrollView);
  }

  // Support global native listeners for scroll events
  private static final Set<ScrollListener> sScrollListeners =
      Collections.newSetFromMap(new WeakHashMap<ScrollListener, Boolean>());

  // If all else fails, this is the hardcoded value in OverScroller.java, in AOSP.
  // The default is defined here (as of this diff):
  // https://android.googlesource.com/platform/frameworks/base/+/ae5bcf23b5f0875e455790d6af387184dbd009c1/core/java/android/widget/OverScroller.java#44
  private static int SMOOTH_SCROLL_DURATION = 250;
  private static boolean mSmoothScrollDurationInitialized = false;

  /** Shared by {@link ReactScrollView} and {@link ReactHorizontalScrollView}. */
  public static void emitScrollEvent(ViewGroup scrollView, float xVelocity, float yVelocity) {
    emitScrollEvent(scrollView, ScrollEventType.SCROLL, xVelocity, yVelocity);
  }

  public static void emitScrollBeginDragEvent(ViewGroup scrollView) {
    emitScrollEvent(scrollView, ScrollEventType.BEGIN_DRAG);
  }

  public static void emitScrollEndDragEvent(
      ViewGroup scrollView, float xVelocity, float yVelocity) {
    emitScrollEvent(scrollView, ScrollEventType.END_DRAG, xVelocity, yVelocity);
  }

  public static void emitScrollMomentumBeginEvent(
      ViewGroup scrollView, int xVelocity, int yVelocity) {
    emitScrollEvent(scrollView, ScrollEventType.MOMENTUM_BEGIN, xVelocity, yVelocity);
  }

  public static void emitScrollMomentumEndEvent(ViewGroup scrollView) {
    emitScrollEvent(scrollView, ScrollEventType.MOMENTUM_END);
  }

  private static void emitScrollEvent(ViewGroup scrollView, ScrollEventType scrollEventType) {
    emitScrollEvent(scrollView, scrollEventType, 0, 0);
  }

  private static void emitScrollEvent(
      ViewGroup scrollView, ScrollEventType scrollEventType, float xVelocity, float yVelocity) {
    View contentView = scrollView.getChildAt(0);

    if (contentView == null) {
      return;
    }

    for (ScrollListener scrollListener : sScrollListeners) {
      scrollListener.onScroll(scrollView, scrollEventType, xVelocity, yVelocity);
    }

    ReactContext reactContext = (ReactContext) scrollView.getContext();
    int surfaceId = UIManagerHelper.getSurfaceId(reactContext);
    UIManagerHelper.getEventDispatcherForReactTag(reactContext, scrollView.getId())
        .dispatchEvent(
            ScrollEvent.obtain(
                surfaceId,
                scrollView.getId(),
                scrollEventType,
                scrollView.getScrollX(),
                scrollView.getScrollY(),
                xVelocity,
                yVelocity,
                contentView.getWidth(),
                contentView.getHeight(),
                scrollView.getWidth(),
                scrollView.getHeight()));
  }

  /** This is only for Java listeners. onLayout events emitted to JS are handled elsewhere. */
  public static void emitLayoutEvent(ViewGroup scrollView) {
    for (ScrollListener scrollListener : sScrollListeners) {
      scrollListener.onLayout(scrollView);
    }
  }

  public static int parseOverScrollMode(String jsOverScrollMode) {
    if (jsOverScrollMode == null || jsOverScrollMode.equals(AUTO)) {
      return View.OVER_SCROLL_IF_CONTENT_SCROLLS;
    } else if (jsOverScrollMode.equals(OVER_SCROLL_ALWAYS)) {
      return View.OVER_SCROLL_ALWAYS;
    } else if (jsOverScrollMode.equals(OVER_SCROLL_NEVER)) {
      return View.OVER_SCROLL_NEVER;
    } else {
      throw new JSApplicationIllegalArgumentException("wrong overScrollMode: " + jsOverScrollMode);
    }
  }

  public static int parseSnapToAlignment(@Nullable String alignment) {
    if (alignment == null) {
      return SNAP_ALIGNMENT_DISABLED;
    } else if ("start".equalsIgnoreCase(alignment)) {
      return SNAP_ALIGNMENT_START;
    } else if ("center".equalsIgnoreCase(alignment)) {
      return SNAP_ALIGNMENT_CENTER;
    } else if ("end".equals(alignment)) {
      return SNAP_ALIGNMENT_END;
    } else {
      throw new JSApplicationIllegalArgumentException("wrong snap alignment value: " + alignment);
    }
  }

  public static int getDefaultScrollAnimationDuration(Context context) {
    if (!mSmoothScrollDurationInitialized) {
      mSmoothScrollDurationInitialized = true;

      try {
        OverScrollerDurationGetter overScrollerDurationGetter =
            new OverScrollerDurationGetter(context);
        SMOOTH_SCROLL_DURATION = overScrollerDurationGetter.getScrollAnimationDuration();
      } catch (Throwable e) {
      }
    }

    return SMOOTH_SCROLL_DURATION;
  }

  private static class OverScrollerDurationGetter extends OverScroller {
    // This is the default in AOSP, hardcoded in OverScroller.java.
    private int mScrollAnimationDuration = 250;

    OverScrollerDurationGetter(Context context) {
      // We call with a null context because OverScroller does not use the context
      // in the execution path we're interested in, unless heavily modified in an AOSP fork.
      super(context);
    }

    public int getScrollAnimationDuration() {
      // If startScroll is called without a duration, OverScroller will call `startScroll(x, y, dx,
      // dy, duration)` with the default duration.
      super.startScroll(0, 0, 0, 0);

      return mScrollAnimationDuration;
    }

    @Override
    public void startScroll(int startX, int startY, int dx, int dy, int duration) {
      mScrollAnimationDuration = duration;
    }
  }

  /**
   * Adds a scroll listener.
   *
   * <p>Note that you must keep a reference to this scroll listener because this class only keeps a
   * weak reference to it (to prevent memory leaks). This means that code like <code>
   * addScrollListener(new ScrollListener() {...})</code> won't work, you need to do this instead:
   * <code>
   *   mScrollListener = new ScrollListener() {...};
   *   ReactScrollViewHelper.addScrollListener(mScrollListener);
   * </code> instead.
   *
   * @param listener
   */
  public static void addScrollListener(ScrollListener listener) {
    sScrollListeners.add(listener);
  }

  public static void removeScrollListener(ScrollListener listener) {
    sScrollListeners.remove(listener);
  }

  public static class ReactScrollViewScrollState {
    private final int mLayoutDirection;
    private final Point mFinalAnimatedPositionScroll = new Point();
    private int mScrollAwayPaddingTop = 0;
    private final Point mLastStateUpdateScroll = new Point(-1, -1);

    public ReactScrollViewScrollState(final int layoutDirection) {
      mLayoutDirection = layoutDirection;
    }

    /**
     * Get the layout direction. Can be either scrollView.LAYOUT_DIRECTION_RTL (1) or
     * scrollView.LAYOUT_DIRECTION_LTR (0). If the value is -1, it means unknown layout.
     */
    public int getLayoutDirection() {
      return mLayoutDirection;
    }

    /** Get the position after current animation is finished */
    public Point getFinalAnimatedPositionScroll() {
      return mFinalAnimatedPositionScroll;
    }

    /** Set the final scroll position after scrolling animation is finished */
    public ReactScrollViewScrollState setFinalAnimatedPositionScroll(
        int finalAnimatedPositionScrollX, int finalAnimatedPositionScrollY) {
      mFinalAnimatedPositionScroll.set(finalAnimatedPositionScrollX, finalAnimatedPositionScrollY);
      return this;
    }

    /** Get the Fabric state of last scroll position */
    public Point getLastStateUpdateScroll() {
      return mLastStateUpdateScroll;
    }

    /** Set the Fabric state of last scroll position */
    public ReactScrollViewScrollState setLastStateUpdateScroll(
        int lastStateUpdateScrollX, int lastStateUpdateScrollY) {
      mLastStateUpdateScroll.set(lastStateUpdateScrollX, lastStateUpdateScrollY);
      return this;
    }

    /** Get the padding on the top for nav bar */
    public int getScrollAwayPaddingTop() {
      return mScrollAwayPaddingTop;
    }

    /** Set the padding on the top for nav bar */
    public ReactScrollViewScrollState setScrollAwayPaddingTop(int scrollAwayPaddingTop) {
      mScrollAwayPaddingTop = scrollAwayPaddingTop;
      return this;
    }
  }

  public static <
          T extends ViewGroup & FabricViewStateManager.HasFabricViewStateManager & HasScrollState>
      void smoothScrollTo(final T scrollView, final int x, final int y) {
    if (DEBUG_MODE) {
      FLog.i(TAG, "smoothScrollTo[%d] x %d y %d", scrollView.getId(), x, y);
    }

    // `smoothScrollTo` contains some logic that, if called multiple times in a short amount of
    // time, will treat all calls as part of the same animation and will not lengthen the duration
    // of the animation. This means that, for example, if the user is scrolling rapidly, multiple
    // pages could be considered part of one animation, causing some page animations to be animated
    // very rapidly - looking like they're not animated at all.
    if (sScrollAnimator != null) {
      sScrollAnimator.cancel();
    }

    final ReactScrollViewScrollState scrollState = scrollView.getReactScrollViewScrollState();
    scrollState.setFinalAnimatedPositionScroll(x, y);
    PropertyValuesHolder scrollX =
        PropertyValuesHolder.ofInt("scrollX", scrollView.getScrollX(), x);
    PropertyValuesHolder scrollY =
        PropertyValuesHolder.ofInt("scrollY", scrollView.getScrollY(), y);
    sScrollAnimator = ObjectAnimator.ofPropertyValuesHolder(scrollX, scrollY);
    sScrollAnimator.setDuration(getDefaultScrollAnimationDuration(scrollView.getContext()));
    sScrollAnimator.addUpdateListener(
        new ValueAnimator.AnimatorUpdateListener() {
          @Override
          public void onAnimationUpdate(ValueAnimator valueAnimator) {
            int scrollValueX = (Integer) valueAnimator.getAnimatedValue("scrollX");
            int scrollValueY = (Integer) valueAnimator.getAnimatedValue("scrollY");
            scrollView.scrollTo(scrollValueX, scrollValueY);
          }
        });
    sScrollAnimator.addListener(
        new Animator.AnimatorListener() {
          @Override
          public void onAnimationStart(Animator animator) {}

          @Override
          public void onAnimationEnd(Animator animator) {
            scrollState.setFinalAnimatedPositionScroll(-1, -1);
            sScrollAnimator = null;
            updateStateOnScroll(scrollView);
          }

          @Override
          public void onAnimationCancel(Animator animator) {}

          @Override
          public void onAnimationRepeat(Animator animator) {}
        });
    sScrollAnimator.start();
    updateStateOnScroll(scrollView, x, y);
  }

  /** Get current (x, y) position or position after current animation finishes, if any. */
  public static <
          T extends ViewGroup & FabricViewStateManager.HasFabricViewStateManager & HasScrollState>
      Point getPostAnimationScroll(final T scrollView) {
    return sScrollAnimator != null && sScrollAnimator.isRunning()
        ? scrollView.getReactScrollViewScrollState().getFinalAnimatedPositionScroll()
        : new Point(scrollView.getScrollX(), scrollView.getScrollY());
  }

  public static <
          T extends ViewGroup & FabricViewStateManager.HasFabricViewStateManager & HasScrollState>
      boolean updateStateOnScroll(final T scrollView) {
    return updateStateOnScroll(scrollView, scrollView.getScrollX(), scrollView.getScrollY());
  }

  /**
   * Called on any stabilized onScroll change to propagate content offset value to a Shadow Node.
   */
  public static <
          T extends ViewGroup & FabricViewStateManager.HasFabricViewStateManager & HasScrollState>
      boolean updateStateOnScroll(final T scrollView, final int scrollX, final int scrollY) {
    if (DEBUG_MODE) {
      FLog.i(
          TAG,
          "updateStateOnScroll[%d] scrollX %d scrollY %d",
          scrollView.getId(),
          scrollX,
          scrollY);
    }

    if (ViewUtil.getUIManagerType(scrollView.getId()) == UIManagerType.DEFAULT) {
      return false;
    }

    final ReactScrollViewScrollState scrollState = scrollView.getReactScrollViewScrollState();
    // Dedupe events to reduce JNI traffic
    if (scrollState.getLastStateUpdateScroll().equals(scrollX, scrollY)) {
      return false;
    }

    scrollState.setLastStateUpdateScroll(scrollX, scrollY);
    forceUpdateState(scrollView);
    return true;
  }

  public static <
          T extends ViewGroup & FabricViewStateManager.HasFabricViewStateManager & HasScrollState>
      void forceUpdateState(final T scrollView) {
    final ReactScrollViewScrollState scrollState = scrollView.getReactScrollViewScrollState();
    final int scrollAwayPaddingTop = scrollState.getScrollAwayPaddingTop();
    final Point scrollPos = scrollState.getLastStateUpdateScroll();
    final int scrollX = scrollPos.x;
    final int scrollY = scrollPos.y;
    final int fabricScrollX;
    int layoutDirection = scrollState.getLayoutDirection();

    if (layoutDirection == scrollView.LAYOUT_DIRECTION_RTL) {
      // getScrollX returns offset from left even when layout direction is RTL.
      // The following line calculates offset from right.
      View child = scrollView.getChildAt(0);
      int contentWidth = child != null ? child.getWidth() : 0;
      fabricScrollX = -(contentWidth - scrollX - scrollView.getWidth());
    } else {
      fabricScrollX = scrollX;
    }

    if (DEBUG_MODE) {
      FLog.i(
          TAG,
          "updateStateOnScroll[%d] scrollX %d scrollY %d fabricScrollX",
          scrollView.getId(),
          scrollX,
          scrollY,
          fabricScrollX);
    }

    scrollView
        .getFabricViewStateManager()
        .setState(
            new FabricViewStateManager.StateUpdateCallback() {
              @Override
              public WritableMap getStateUpdate() {
                WritableMap map = new WritableNativeMap();
                map.putDouble(CONTENT_OFFSET_LEFT, PixelUtil.toDIPFromPixel(scrollX));
                map.putDouble(CONTENT_OFFSET_TOP, PixelUtil.toDIPFromPixel(scrollY));
                map.putDouble(
                    SCROLL_AWAY_PADDING_TOP, PixelUtil.toDIPFromPixel(scrollAwayPaddingTop));
                return map;
              }
            });
  }

  public interface HasScrollState {
    /** Get the scroll state for the current ScrollView */
    ReactScrollViewScrollState getReactScrollViewScrollState();
  }
}
